include string.e


global constant
	LIST_MAIN = 2,	-- Index of the main part of the list (before |)
	LIST_LOOP = 3,	-- Index of the loop part of the list (after |)
	LIST_POS = 3,	-- Index of the list position data
	LIST_RET = 3,	-- Index of the returned data from step_list
	LIST_VAL = 3	-- Index of the value in the returned data from step_list
	
	
-- Read a {}-enclosed list of numbers separated by spaces or commas. The list can
-- contain a loop, beginning at | and ranging to the end of the list.
global function get_list()
	sequence s,t
	integer c, commaOk, endOk, pipeOk, concatTo
	object o
	atom startVal, stopVal, stepVal
	
	-- s[2] will contain everything before the |, s[3] everything after it.
	s = {GET_FAIL, {}, {}}
	
	commaOk = 0		-- Not ok to read a comma
	pipeOk = 1		-- Ok to read a |
	endOk = 0		-- Not ok to read a }
	concatTo = LIST_MAIN	-- Concatenate to s[2]
	
	skip_whitespace()
	
	c = getch()
	
	if c = '{' then
		while 1 do
			skip_whitespace()
			--t = get_string_in_range(DIGITS & '-')
			t = get_numeric_string()
			if length(t) then
				o = value(t)
				if o[1] = GET_SUCCESS then
					skip_whitespace()
					c = getch()
					if c = ':' then
						startVal = floor(o[2])
						t = get_numeric_string()
						o = value(t)
						if o[1] = GET_SUCCESS then
							stopVal = floor(o[2])
							skip_whitespace()
							c = getch()
							if c = ':' then
								allow_floats_in_numeric_strings()
								t = get_numeric_string()
								o = value(t)
								if o[1] = GET_SUCCESS then
									stepVal = o[2]
								else
									ERROR("Malformed interval: " & t, lineNum)
								end if
							else
								ungetch()
								if stopVal < startVal then
									stepVal = -1
								else
									stepVal = 1
								end if
							end if
						else
							ERROR("Malformed interval: " & t, lineNum)
						end if
						if (stopVal < startVal and stepVal >= 0) then
							ERROR("Bad step value for interval", lineNum)
						elsif (stopVal > startVal and stepVal <= 0) then
							ERROR("Bad step value for interval", lineNum)
						end if
						--if (stopVal < startVal) then
						--	while startVal >= stopVal do
						--		s[concatTo] &= floor(startVal)
						--		startVal += stepVal
						--	end while
						--else
						--	while startVal <= stopVal do
						--		s[concatTo] &= floor(startVal)
						--		startVal += stepVal
						--	end while
						--end if
						for i = startVal to stopVal by stepVal do
							s[concatTo] &= floor(i)
						end for
					else
						ungetch()
						s[concatTo] &= o[2]
					end if
					commaOk = 1
					pipeOk = 1
					endOk = 1
				else
					ERROR("Syntax error: " & t, lineNum)
				end if
			else
				c = getch()
				if c = ',' then
					if not commaOk then
						ERROR("Unexpected comma", lineNum)
					end if
					commaOk = 0
					pipeOk = 0
					endOk = 0
				elsif c = '|' then
					if concatTo = LIST_MAIN and pipeOk then
						concatTo = LIST_LOOP
						commaOk = 0
						pipeOk = 0
						endOk = 0
					else
						ERROR("Unexpected |", lineNum)
					end if
				elsif c = '}' then
					if endOk then
						s[1] = GET_SUCCESS
						exit
					else
						ERROR("Malformed list", lineNum)
					end if
				elsif c = ';' then
					while c != 10 and c != -1 do
						c = getch()
					end while
				elsif c = '\"' then
					t = ""
					c = getch()
					while c != '\"' and c != -1 do
						t &= c
						c = getch()
					end while
					s[concatTo] &= {t}
				elsif c = -1 then
					exit
				end if
			end if
		end while
	else
		ERROR("Expected {, got " & c, lineNum)
	end if
	
	if s[1] = GET_SUCCESS then
		skip_whitespace()
		c = getch()
		if c = '+' or c = '-' or c = '*' then
			skip_whitespace()
			allow_floats_in_numeric_strings()
			t = get_numeric_string()
			if length(t) then
				o = value(t)
				if o[1] = GET_SUCCESS then
					if c = '+' then
						s[2] = floor(s[2] + o[2])
						s[3] = floor(s[3] + o[2])
					elsif c = '-' then
						s[2] = floor(s[2] - o[2])
						s[3] = floor(s[3] - o[2])
					elsif c = '*' then
						s[2] = floor(s[2] * o[2])
						s[3] = floor(s[3] * o[2])
					end if
				else
					ERROR("Syntax error: " & t, lineNum)
				end if
			else
				ERROR("Expected a numeric constant after " & c, lineNum)
			end if
		else
			ungetch()
		end if
	end if
	
	return s
end function


-- Step through a list
-- Return {new_position next_item}
global function step_list(sequence l, sequence pos)
	sequence t
	if pos[1] <= length(l[pos[2]]) then
		t = {l[pos[2]][pos[1]]}
		pos[1] += 1
		t = pos[1..2] & t
	else 
		if length(l[LIST_LOOP]) then
			pos[2] = LIST_LOOP
			pos[1] = 1
			t = {l[pos[2]][pos[1]]}
			pos[1] += 1
			t = pos[1..2] & t
		else
			t = {length(l[2]), 2, l[2][length(l[2])]}
		end if
	end if
	
	return t
end function
		
	
-- Return a sequence containing a text-form representation on a list returned by get_list
-- 
-- Examples:
--   {GET_SUCCESS, {1, 2, 3}, {}} -> "{1 2 3}"
--   {GET_SUCCESS, {1, 2, 3}, {4, 5}} -> "{1 2 3 | 4 5}"
--   {GET_SUCCESS, {}, {4, 5}} -> "{| 4 5}"
--
global function sprint_list(sequence l)
	sequence t
	
	t = "{"
	for i = 1 to length(l[LIST_MAIN]) do
		if sequence(l[LIST_MAIN][i]) then
			t &= l[LIST_MAIN][i] & " "
		else
			t &= sprintf("%d", l[LIST_MAIN][i])
		end if
		if i < length(l[LIST_MAIN]) then
			t &= ' '
		end if
	end for
	
	if length(l[LIST_LOOP]) then
		t &= " | "
		for i = 1 to length(l[LIST_LOOP]) do
			t &= sprintf("%d", l[LIST_LOOP][i])
			if i < length(l[LIST_LOOP]) then
				t &= ' '
			end if
		end for
	end if
	
	return t & '}'
end function


global function is_empty_list(sequence l)
	return (length(l[LIST_MAIN]) + length(l[LIST_LOOP])) = 0
end function
